
#include "version.h"

#define STR2(x)                 #x
#define STR(x)                  STR2(x)

#define APPNAME                 "PNR"
#define APPANDVER               APPNAME " " STR(PVERSION)
#define APPANDVERANDDATE        APPANDVER " " __DATE__
#define APPFULLMONTY            APPANDVERANDDATE \
                                " - Pick 'n' Run by ChAoS@glod.net"

/*------------------------------------------------------------------------*/

/*

20020620 MSD 013     Added the -DA, -DC and -DM options.  These make the
                     dialogue box, child window and menu positions,
                     respectively, follow the current mouse position.

20020609 MSD 012     If the pick box, for whatever reason, isn't displayed,
                     then any following menu can't be used from the
                     keyboard. This has been fixed.

20010908 MSD 011 (a) Added %PARAMn% facility.

                 (b) Removed the DEFWINDOWPROC() calls from within *dialogue*
                     procedures because the WIN32 API help file says that you
                     shouldn't *do* this. (I don't know *why* this is
                     considered a Bad Thing.)

                 (c) Fixed the bug whereby *two* functions (EXPANDSTR() and
                     EXPANDENVALL()) thought that they were responsible for
                     reducing "escaped" percent signs.  As a result,
                     multiple percents were sometimes required in order to
                     get a *single* percent to processed.

20010526 MSD 010     Altered to send all diagnostic output to the standard
                     Windows debug output stream instead of using a myriad of
                     (annoying) dialogue boxes.

20010508 MSD 009 (a) Added the ability to *create* files via the open file
                     dialogue box.  (This is the "-DC" option.)

                 (b) Added the use of PNR's very own icon.

20010501 MSD 008     Increased MAX_MENU_ITEMS from 20 to 40.

20010329 MSD 007 (a) Added the new menuing system that allows the user to
                     pick a file and then pick a *command-line* from a menu
                     generated on the fly.

                 (b) Relaxed the use of white-space in a response file.
                     Now, all white space characters (well, tab, CR and LF,
                     at least) are translated into spaces before inserting
                     them into the Windows command-line.

20010315 MSD 006     A bit of fiddling with the %4 and %5 parameter handling
                     in the case of multi-select, single command-lines.

20010314 MSD 005 (a) Added the multi-select single-command facility. In the
                     past, when a multi-select was done, PNR would start a
                     separate command shell for *each* file slected.  This
                     new alternative allows you to start a single command
                     shell placing *all* the selected files in one "%1"
                     parameter. Hence, for example, one can select several
                     files and have them all presented by *one*
                     instantiation of your editor.  (Some editors, of
                     course, will detect this and do the Right Thing.
                     However, an MSdos editor run under Win95, for instance,
                     probably won't.)
                 (b) Added AUTODIR facility.
                 (c) Various little cosmetic changes.
                 (d) Removed all modification tags because these changes were
                     extensive enough that I forgot to tag some lines...

20010313 MSD 004 (a) Added auto-quote code.
                 (b) Made some local parameters static.
                 (c) In BOXTXT(), the local char string TEXT was too small - it
                     was used as a SPRINTF() target when the source could be
                     *very* long.  (It's now a *large* STATIC string.)
                 (d) Add the ability to set, reset and toggle switches.  For
                     example, as well as "-CQ" one can now use "-CQ+" and
                     "-CQ-".

20010311 MSD 003 (a) Adopt new version number scheme.
                 (b) Fix path with no return value in SETENVVAR().
                 (c) Added -CH parameter that forces *short* file names to
                     be sent to the child process.
                 (d) Increase maximum command-line length from 1024 to 4096.

20000725 MSD 002     Added call to allow the default dialogue procedure to
                     handle all key-board presses in the help "window". The
                     dialogue only now terminates if one of the two buttons
                     is pressed.

20000724 MSD 001 (a) Added OFN_HIDEREADONLY to the OFN.FLAGS field of the
                     pick box so that the "Open as read-only" check box is
                     inhibited.

                 (b) Added this change log.

                 (c) Added www.glod.net to help page and to "full Monty"
                     #DEFINEd text.  (Neither of these changes are flagged.)

*/

/*------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <io.h>
#include <shellapi.h>
#include "pnr.h"

/*------------------------------------------------------------------------*/

#define PARAMS                          ( 10 )

#define MAX_COMMAND_LINE_LENGTH         ( 8 * 1024 )

#define MAX_MENU_CMD_LEN                ( 1024 )
#define MAX_MENU_NAM_LEN                ( 64 )
#define MAX_MENU_ITEMS                  ( 40 )

#define MENU_PERIOD                     ( 1000 )

#define EXIT_CODE_SUCCESS               ( 0 )
#define EXIT_CODE_BAD_SELECTION         ( 1 )
#define EXIT_CODE_BAD_EXECUTE           ( 2 )
#define EXIT_CODE_BAD_DIR               ( 3 )
#define EXIT_CODE_BAD_FOPEN             ( 4 )
#define EXIT_CODE_BAD_PARAMS            ( 5 )
#define EXIT_CODE_NO_COMSPEC            ( 6 )
#define EXIT_CODE_BAD_EXITPROC          ( 7 )
#define EXIT_CODE_BAD_WINDOW            ( 8 )
#define EXIT_CODE_BAD_DESTROY           ( 9 )
#define EXIT_CODE_BAD_MODFILENAME       ( 10 )
#define EXIT_CODE_BAD_RESPONSE          ( 11 )
#define EXIT_CODE_BAD_FILEFIND          ( 12 )
#define EXIT_CODE_BAD_CONSOLE           ( 13 )
#define EXIT_CODE_BAD_CONRESIZE         ( 14 )
#define EXIT_CODE_BAD_CONGETSIZE        ( 15 )
#define EXIT_CODE_BAD_SHORTNAME         ( 16 )
#define EXIT_CODE_BAD_MENU              ( 17 )
#define EXIT_CODE_PARAM_CANCEL          ( 18 )

/*------------------------------------------------------------------------*/

#define ZERO                    ( 0 )
#define ONE                     ( 1 )
#define TWO                     ( 2 )
#define THREE                   ( 3 )

#define ROGUE_INT               ( -1 )

#define NILSTR                  ( "" )
#define SPACESTR                ( " " )
#define DOTSTR                  ( "." )
#define BACKSLASHSTR            ( "\\" )
#define DQUOTESTR               ( "\"" )

#define EOS                     ( '\0' )
#define TAB                     ( '\t' )
#define CARRIAGERETURN          ( '\r' )
#define LINEFEED                ( '\n' )
#define SPACE                   ( ' ' )
#define MINUSSIGN               ( '-' )
#define PLUSSIGN                ( '+' )
#define SLASH                   ( '/' )
#define SINGLEQUOTE             ( '\'' )
#define BACKSLASH               ( '\\' )
#define DOUBLEQUOTE             ( '"' )
#define PERCENT                 ( '%' )
#define COLON                   ( ':' )
#define COMMA                   ( ',' )
#define EQUAL                   ( '=' )
#define EXCLAMATION             ( '!' )
#define ATSIGN                  ( '@' )
#define DOT                     ( '.' )

#define COMSPEC                 "COMSPEC"
#define PARAM                   "PARAM"

/*------------------------------------------------------------------------*/

typedef struct  _MenuLine
{
        char    Name[MAX_MENU_NAM_LEN+ONE];
        char    Work[MAX_MENU_CMD_LEN+ONE];
}       MenuLine;

static  MenuLine MenuItem[MAX_MENU_ITEMS+ONE];

static  MenuItems = ZERO;

/*------------------------------------------------------------------------*/

static  HINSTANCE Inst;
static  HWND    Root = NULL;

static  char    FileMask[MAX_COMMAND_LINE_LENGTH+ONE];
static  char    OFileName[MAX_PATH+ONE];
static  char    Title[MAX_PATH+ONE];
static  char    StartDir[MAX_PATH+ONE];

static  char    Filters[MAX_COMMAND_LINE_LENGTH+ONE] = { "\0" };
static  int     FiltersLast = ZERO;

static  int     ParamThis   = ZERO;
static  char    ParamText[PARAMS][MAX_PATH+ONE];

static  BOOL    EchoOrders  = FALSE;
static  BOOL    RedirectSTD = FALSE;
static  BOOL    MakeConsole = FALSE;
static  BOOL    ForceShell  = FALSE;
static  BOOL    SplitCmd    = FALSE;
static  BOOL    MakeVisible = FALSE;
static  BOOL    PickSingle  = FALSE;
static  BOOL    WantHelp    = TRUE;
static  BOOL    WantMulti   = FALSE;
static  BOOL    WantRepeat  = FALSE;
static  BOOL    InhibitPick = FALSE;
static  BOOL    WantExp     = FALSE;
static  BOOL    WantCreate  = FALSE;
static  BOOL    Debugging   = FALSE;
static  BOOL    PassEnvVar  = FALSE;
static  BOOL    ShortFNames = FALSE;
static  BOOL    OneMultiCmd = FALSE;
static  BOOL    AutoQuote   = TRUE;
static  BOOL    AutoDir     = TRUE;

static  int     WaitTilDone = ROGUE_INT;
static  int     LaunchGap   = ROGUE_INT;

static  int     BaseCmd     = ROGUE_INT;
static  int     PickCmd     = ROGUE_INT;
static  BOOL    Abandoned   = FALSE;
static  int     TimerHandle;

static  int     DPosX       = ROGUE_INT;
static  int     DPosY       = ROGUE_INT;

static  int     MPosX       = ROGUE_INT;
static  int     MPosY       = ROGUE_INT;

static  int     PosX        = ROGUE_INT;
static  int     PosY        = ROGUE_INT;

static  int     SizX        = ROGUE_INT;
static  int     SizY        = ROGUE_INT;

static  BOOL    DFollow     = FALSE;
static  BOOL    CFollow     = FALSE;
static  BOOL    MFollow     = FALSE;

enum    { SS_NONE, SS_MAX, SS_MIN, SS_RES } ShowState = SS_NONE;

/*========================================================================*/

static  char GenText[MAX_COMMAND_LINE_LENGTH+ONE];

/*------------------------------------------------------------------------*/

void    xprintf(        char            *fmt,
                        ... )
{
static  char    Message[128];

        va_list ap;

        va_start(ap,fmt);
        vsprintf(GenText,fmt,ap);
        va_end(ap);

        OutputDebugString(GenText);
}

/*------------------------------------------------------------------------*/

static  int GLEnumber;

void    Box(            char    *message )
{
        GLEnumber = GetLastError();

        MessageBox(NULL,message,APPANDVERANDDATE,MB_OK);
}


void    BoxTxt(         char    *message,
                        char    *txt )
{
        GLEnumber = GetLastError();

        sprintf(GenText,"%s: <%s>",message, txt==NULL ? "<null>" : txt);

        MessageBox(NULL,GenText,APPANDVERANDDATE,MB_OK);
}


void    BoxInt(         char    *message,
                        int     value )
{
        GLEnumber = GetLastError();

        sprintf(GenText,"%s: %d = 0x%X",message,value,value);

        MessageBox(NULL,GenText,APPANDVERANDDATE,MB_OK);
}


void    BoxErrNo(       int     ErrNo )
{
        sprintf(GenText,"Error number: %d = 0x%X",ErrNo,ErrNo);

        MessageBox(NULL,GenText,APPANDVERANDDATE,MB_OK);
}

/*      You must have previously called BOX(), BOXTXT() or BOXINT() so that
 *      GLENUMBER has already been set up with the value from GETLASTERROR().
 */

void    BoxErrGLE(      void )
{
        BoxInt("GetLastError",GLEnumber);
}

/*------------------------------------------------------------------------*/

void    DbgBox(         char    *message )
{
        if ( !Debugging ) return;

        GLEnumber = GetLastError();

        xprintf("PNR: %s\n",message);
}


void    DbgBoxTxt(      char    *message,
                        char    *txt )
{
        if ( !Debugging ) return;

        GLEnumber = GetLastError();

        xprintf("PNR: %s: <%s>\n",message, txt==NULL ? "<null>" : txt);
}


void    DbgBoxInt(      char    *message,
                        int     value )
{
        if ( !Debugging ) return;

        GLEnumber = GetLastError();

        xprintf("PNR: %s: %d = 0x%X\n",message,value,value);
}


void    DbgBoxErrNo(    int     ErrNo )
{
        if ( !Debugging ) return;

        xprintf("PNR: Error number: %d = 0x%X\n",ErrNo,ErrNo);
}

void    DbgBoxErrGLE(   void )
{
        DbgBoxInt("GetLastError",GLEnumber);
}

/*========================================================================*/

BOOL    IsPresent(      char    *s )
{
        return ( strlen(s) > ZERO );
}

/*------------------------------------------------------------------------*/

BOOL    IsSame(         char    *s,
                        char    *t )
{
        return  ( strcmp(s,t) == ZERO );
}

/*------------------------------------------------------------------------*/

/* Returns position of char C in string S.  Returns -1 if it can't find C. */

int     strpos(         char    *s,
                        char    c )
{
        int i;
        int len = strlen(s);

        for     ( i=ZERO ; i<len ; i++ )
                if ( s[i] == c ) return i;

        return -ONE;
}

/*------------------------------------------------------------------------*/

/* Replaces all occurrences of OLD with NEW in string S.  It returns the
 * number of replacesments made.
 */

int     strrep(         char    *s,
                        char    Old,
                        char    New )
{
        int Replacements = ZERO;

        while   ( *s )
        {
                if      ( *s == Old )
                {
                        *s = New;
                        Replacements++;
                }

                s++;
        }

        return  Replacements;
}

/*------------------------------------------------------------------------*/

/* Truncates string S at the first occurence (if any) of the character C. */

void    strtrunc(       char    *s,
                        char    c )
{
        int i = strpos(s,c);

        if ( i >= ZERO ) s[i] = EOS;
}

/*------------------------------------------------------------------------*/

char    *strdel(        char    *s,
                        int     pos,
                        int     num )
{
        int     len = strlen(s);

        if      ( ( len > 0 ) && ( num > 0 ) && ( pos < len ) )
                if      ( num > ( len - pos + 1 ) )     /* Then deleting to */
                        s[pos] = EOS;                   /* the end of "s".  */
                else
                        strcpy(s+pos,s+pos+num);        /* Else just copy   */
                                                        /* the rest down.   */
        return  ( s );
}

/*------------------------------------------------------------------------*/

char    *strins(        char    *targ,
                        char    *src,
                        int     pos )
{
        char    *s;
        char    *t;
        int     lenT;
        int     lenS;
        int     i;

        lenT = strlen(targ), lenS = strlen(src);

        if ( ( pos < 0 ) || ( lenS <= 0 ) ) return targ;

        if      ( pos >= lenT )
        {
                strcat(targ,src);
                return targ;
        }

        s = targ + lenT;
        t = s + lenS;
        for( i=0; i<(lenT-pos+1); ++i ) *t-- = *s--;

        s = src;
        t = targ + pos;
        for ( i=0; i<lenS; ++i ) *t++ = *s++;

        return  ( targ );
}

/*------------------------------------------------------------------------*/

void    StripCRLF(      char    *s )
{
        while   ( IsPresent(s)
        &&        ( ( s[strlen(s)-ONE] == CARRIAGERETURN )
          ||        ( s[strlen(s)-ONE] == LINEFEED ) ) )
                strdel(s,strlen(s)-ONE,ONE);
}

/*------------------------------------------------------------------------*/

void    StripTrailChar( char    *s,
                        char    c )
{
        while   ( IsPresent(s) && ( s[strlen(s)-ONE] == c ) )
                strdel(s,strlen(s)-ONE,ONE);
}

/*------------------------------------------------------------------------*/

void    StripLeadChar(  char    *s,
                        char    c )
{
        while ( s[ZERO] == c ) strdel(s,ZERO,ONE);
}

/*------------------------------------------------------------------------*/

void    ParseFileName(  char    *FileName,
                        char    *Drive,
                        char    *Path,
                        char    *Root,
                        char    *Ext )
{
        char    *StartOfPath, *StartOfRoot, *StartOfExt;

        *Drive = *Path = *Root = *Ext = EOS;

        if      ( isalpha(FileName[ZERO]) && FileName[ONE] == COLON )
        {
                Drive[ZERO] = FileName[ZERO];
                Drive[ONE]  = FileName[ONE];
                Drive[TWO]  = EOS;

                StartOfPath = FileName + TWO;
        }
        else
                StartOfPath = FileName;

        if      ( (StartOfRoot=strrchr(StartOfPath,BACKSLASH)) != NULL )
                StartOfRoot++;
        else
                StartOfRoot = StartOfPath;

        if      ( (StartOfExt=strrchr(StartOfRoot,DOT)) != NULL )
                StartOfExt++;
        else
                StartOfExt = StartOfRoot + strlen(StartOfRoot);

        strncpy(Path,StartOfPath,StartOfRoot-StartOfPath);
        Path[StartOfRoot-StartOfPath] = EOS;
        StripTrailChar(Path,BACKSLASH);

        strncpy(Root,StartOfRoot,StartOfExt-StartOfRoot);
        Root[StartOfExt-StartOfRoot] = EOS;
        StripTrailChar(Root,DOT);

        strcpy(Ext,StartOfExt);
}

/*------------------------------------------------------------------------*/

BOOL    QuoteStr(       char    *s )
{
        DbgBoxTxt("QUOTESTR(s) s",s);

        if ( strpos(s,SPACE) < ZERO ) return FALSE;

        strins(s,DQUOTESTR,ZERO);
        strins(s,DQUOTESTR,strlen(s));

        DbgBoxTxt("QUOTESTR(e) s",s);

        return TRUE;
}

/*------------------------------------------------------------------------*/

static  char    *DigitNames[] =
                {
                        "Zero", "One", "Two", "Three", "Four", "Five",
                        "Six", "Seven", "Eight", "Nine"
                };

BOOL
CALLBACK DlgProcParamText(HWND  hwnd,
                        UINT    iMsg,
                        WPARAM  wParam,
                        LPARAM  lParam)
{
static  char    Desc[64];

        switch  ( iMsg )
        {
        case WM_INITDIALOG:
                SetWindowText(hwnd,APPFULLMONTY);
                sprintf(Desc,"Enter Parameter %s",DigitNames[ParamThis]);
                SetDlgItemText(hwnd,IDC_PARAM_DESC,Desc);
                return TRUE;

        case WM_COMMAND:
                if      ( HIWORD(wParam) == BN_CLICKED )
                        switch  ( (int)LOWORD(wParam) )
                        {
                        case IDOK:
                                GetDlgItemText(hwnd,IDC_PARAM_TEXT,
                                       ParamText[ParamThis],
                                       sizeof(ParamText[ZERO])-ONE);
                                EndDialog(hwnd,TRUE);
                                return TRUE;

                        case IDCANCEL:
                                EndDialog(hwnd,FALSE);
                                return TRUE;
                        }
        }

        return FALSE;
}

/*------------------------------------------------------------------------*/

char    *ExpandStr(     char    *s,
                        char    *P1,
                        char    *P2,
                        char    *P3,
                        char    *P4,
                        char    *P5 )
{
        char    *start = s;

        while   ( *start )
                if      ( *start != PERCENT )
                        start++;
                else
                {
                        strdel(start,ZERO,ONE);

                        switch  ( *start )
                        {
                        case '1':       strdel(start,ZERO,ONE);
                                        strins(start,P1,ZERO);
                                        start += strlen(P1);
                                        break;

                        case '2':       strdel(start,ZERO,ONE);
                                        strins(start,P2,ZERO);
                                        start += strlen(P2);
                                        break;

                        case '3':       strdel(start,ZERO,ONE);
                                        strins(start,P3,ZERO);
                                        start += strlen(P3);
                                        break;

                        case '4':       strdel(start,ZERO,ONE);
                                        strins(start,P4,ZERO);
                                        start += strlen(P4);
                                        break;

                        case '5':       strdel(start,ZERO,ONE);
                                        strins(start,P5,ZERO);
                                        start += strlen(P5);
                                        break;

                        default:        start++;
                                        break;
                        }
                }

        return s;
}

/*========================================================================*/

void    AllToUpper(     char            *s )
{
        while   ( *s )
        {
                if ( islower(*s) ) *s = toupper(*s);
                s++;
        }
}

/*------------------------------------------------------------------------*/

BOOL    DoParamText(    void )
{
        if ( IsPresent(ParamText[ParamThis]) ) return TRUE;

        return DialogBox(Inst,MAKEINTRESOURCE(IDD_PARAM),Root,
                                                     (DLGPROC)DlgProcParamText);
}

/*------------------------------------------------------------------------*/

char    *ExpandEnvOne(  char    *Start,
                        char    *End )
{
static  char    ParName[MAX_PATH+ONE];
static  char    VarName[MAX_PATH+ONE];
static  char    VarVal[MAX_COMMAND_LINE_LENGTH+ONE];

        strcpy(VarName,Start+ONE); VarName[End-Start-ONE] = EOS;

        strcpy(ParName,VarName);
        AllToUpper(ParName);
        if   ( ( strlen(ParName) == ( strlen(PARAM) + ONE ) )
        &&     ( strncmp(ParName,PARAM,strlen(PARAM)) == ZERO )
        &&     isdigit(ParName[strlen(PARAM)]) )
        {
                strdel(Start,ZERO,End-Start+ONE);
                ParamThis = ParName[strlen(PARAM)] - '0';
                if ( !DoParamText() ) return NULL;
                strins(Start,ParamText[ParamThis],ZERO);
                return Start;
        }

        if   ( (GetEnvironmentVariable(VarName,VarVal,sizeof(VarVal))) == ZERO )
             return Start + ONE;

        strdel(Start,ZERO,End-Start+ONE);
        strins(Start,VarVal,ZERO);
        return Start;
}

/*------------------------------------------------------------------------*/

char    *ExpandEnvAll(  char    *s )
{
        char    *start = s;
        int     pos;

        while   ( *start )
        {
                if      ( *start != PERCENT )
                {
                        start++;
                        continue;
                }

                if      ( (pos=strpos(start+ONE,PERCENT)) < ZERO )
                        return s;

                start = ExpandEnvOne(start,start+pos+ONE);

                if ( start == NULL ) return NULL;
        }

        return s;
}

/*========================================================================*/

void    Get2Nums(       char    *s,
                        int     *x,
                        int     *y )
{
        int Pos = strpos(s,COMMA);

        *x = atoi(s), *y = ZERO;

        if ( Pos > ZERO ) *y = atoi(s+Pos+ONE);
}

/*------------------------------------------------------------------------*/

BOOL    AddFilter(      char    *NewFilters )
{
static  char    FilterCpy[MAX_COMMAND_LINE_LENGTH+ONE];
        char    *s = FilterCpy;
        int     Pos;

        strcpy(FilterCpy,NewFilters);

        FilterCpy[strlen(FilterCpy)+ONE] = EOS;

        while   ( IsPresent(s) )
        {
                Pos = strpos(s,EXCLAMATION);
                if ( Pos < ZERO ) Pos = strlen(s);
                s[Pos] = EOS;

                StripLeadChar(s,SPACE);
                StripTrailChar(s,SPACE);

                if      ( FiltersLast + strlen(s) >= sizeof(Filters) )
                {
                        BoxTxt("Too much filter data",s);
                        return FALSE;
                }

                strcpy(Filters+FiltersLast,s);

                FiltersLast += strlen(s) + ONE;
                Filters[FiltersLast] = EOS;

                s += Pos + ONE;
        }

        return TRUE;
}

/*------------------------------------------------------------------------*/

BOOLEAN SetEnvVar(      char    *Parm )
{
static  char    VarVal[MAX_COMMAND_LINE_LENGTH+ONE];
        int     Pos;

        if ( (Pos=strpos(Parm,EQUAL)) < ZERO ) return FALSE;

        Parm[Pos] = EOS;

        strcpy(VarVal,Parm+Pos+ONE);

        if ( ExpandEnvAll(VarVal) == NULL ) return FALSE;

        if      ( !SetEnvironmentVariable(Parm,VarVal) )
        {
                BoxTxt("Cannot set environment variable",Parm);
                BoxErrGLE();
                return FALSE;
        }

        return TRUE;
}

/*------------------------------------------------------------------------*/

/* Handles parameter strings like "-X", "-X+" or "-X-".  The trailing
 * character is the one to pass down.  If it's a plus sign, the switch is set,
 * if it's a munis sign the switch is reset other wise it's toggled.
 */

void    SetSwitch(      BOOL    *Switch,
                        char    Terminator )
{
        if      ( Terminator == PLUSSIGN )
                *Switch = TRUE;
        else
        if      ( Terminator == MINUSSIGN )
                *Switch = FALSE;
        else
                *Switch = !(*Switch);
}

/*------------------------------------------------------------------------*/

BOOLEAN ArgDialog(      char    *Parm )
{
        BOOL Ok = TRUE;

        switch  ( toupper(*Parm++) )
        {
        case 'A':       SetSwitch(&DFollow,*Parm);                      break;
        case 'C':       SetSwitch(&WantCreate,*Parm);                   break;
        case 'E':       SetSwitch(&WantExp,*Parm);                      break;
        case 'F':       Ok = AddFilter(Parm);                           break;
        case 'H':       SetSwitch(&WantHelp,*Parm);                     break;
        case 'I':       SetSwitch(&MakeVisible,*Parm);                  break;
        case 'M':       strcpy(FileMask,Parm);                          break;
        case 'O':       strcpy(OFileName,Parm);                         break;
        case 'P':       Get2Nums(Parm,&DPosX,&DPosY);                   break;
        case 'R':       SetSwitch(&WantRepeat,*Parm);                   break;
        case 'S':       SetSwitch(&PickSingle,*Parm);                   break;
        case 'T':       strcpy(Title,Parm);                             break;
        case 'U':       SetSwitch(&WantMulti,*Parm);                    break;
        case 'V':       Ok = SetEnvVar(Parm);                           break;
        case 'W':       LaunchGap = atoi(Parm);                         break;
        case 'X':       SetSwitch(&InhibitPick,*Parm);                  break;
        default:        Ok = FALSE;                                     break;
        }

        return Ok;
}

/*------------------------------------------------------------------------*/

BOOLEAN ArgChild(       char    *Parm )
{
        BOOL Ok = TRUE;

        switch  ( toupper(*Parm++) )
        {
        case 'A':       SetSwitch(&CFollow,*Parm);                      break;
        case 'C':       SetSwitch(&SplitCmd,*Parm);                     break;
        case 'D':       strcpy(StartDir,Parm);                          break;
        case 'E':       SetSwitch(&EchoOrders,*Parm);                   break;
        case 'H':       SetSwitch(&ShortFNames,*Parm);                  break;
        case 'G':       SetSwitch(&OneMultiCmd,*Parm);                  break;
        case 'I':       SetSwitch(&RedirectSTD,*Parm);                  break;
        case 'M':       ShowState = SS_MAX;                             break;
        case 'N':       ShowState = SS_MIN;                             break;
        case 'O':       SetSwitch(&MakeConsole,*Parm);                  break;
        case 'P':       Get2Nums(Parm,&PosX,&PosY);                     break;
        case 'Q':       SetSwitch(&AutoQuote,*Parm);                    break;
        case 'R':       ShowState = SS_RES;                             break;
        case 'S':       SetSwitch(&ForceShell,*Parm);                   break;
        case 'U':       WaitTilDone = atoi(Parm);                       break;
        case 'V':       PassEnvVar = TRUE;
                        Ok = SetEnvVar(Parm);                           break;
        case 'Y':       SetSwitch(&AutoDir,*Parm);                      break;
        case 'Z':       Get2Nums(Parm,&SizX,&SizY);                     break;
        default:        Ok = FALSE;                                     break;
        }

        return Ok;
}

/*------------------------------------------------------------------------*/

BOOLEAN ArgMenu(        char    *Parm )
{
        BOOL Ok = TRUE;

        switch  ( toupper(*Parm++) )
        {
        case 'A':       SetSwitch(&MFollow,*Parm);                      break;

        case 'C':       strcpy(MenuItem[MenuItems].Work,Parm);
                        if      ( IsPresent(MenuItem[MenuItems].Name)
                        &&        ( MenuItems < MAX_MENU_ITEMS ) )
                                MenuItems++;
                                                                        break;
        case 'N':       strcpy(MenuItem[MenuItems].Name,Parm);
                        if      ( IsPresent(MenuItem[MenuItems].Work)
                        &&        ( MenuItems < MAX_MENU_ITEMS ) )
                                MenuItems++;
                                                                        break;
        case 'B':       BaseCmd = atoi(Parm);                           break;
        case 'P':       Get2Nums(Parm,&MPosX,&MPosY);                   break;
        default:        Ok = FALSE;                                     break;
        }

        return Ok;
}

/*------------------------------------------------------------------------*/

BOOLEAN DoParam(        char    *Parm )
{
        BOOL ArgOk = TRUE;

        if      ( strlen(Parm) < THREE )
                ArgOk = FALSE;
        else
                switch  ( toupper(Parm[ONE]) )
                {
                case 'D': ArgOk = ArgDialog(Parm+TWO);                  break;

                case 'C': ArgOk = ArgChild(Parm+TWO);                   break;

                case 'M': ArgOk = ArgMenu(Parm+TWO);                    break;

                case 'X': if    ( toupper(Parm[TWO]) == 'X' )
                                SetSwitch(&Debugging,Parm[THREE]);
                          else
                                ArgOk = FALSE;
                          break;

                default:  ArgOk = FALSE;
                }

        if ( !ArgOk ) BoxTxt("Bad parameter",Parm);

        return ArgOk;
}

/*------------------------------------------------------------------------*/

BOOLEAN ParamStart(     char    *Cmd )
{
        return ( ( *Cmd == MINUSSIGN ) || ( *Cmd == SLASH ) );
}

BOOLEAN IsParam(        char    *Cmd )
{
        if ( ParamStart(Cmd) ) return TRUE;

        return ( ( *Cmd == DOUBLEQUOTE ) || ( *Cmd == SINGLEQUOTE ) )
        &&      ParamStart(Cmd+ONE);
}

/*------------------------------------------------------------------------*/

BOOLEAN GetFileText(    char    *Name,
                        char    *Data )
{
        int     Next;
        FILE    *fd;

        if ( (fd=fopen(Name,"r")) == NULL ) return FALSE;

        while   ( (Next=fgetc(fd)) != EOF )
        {
                if      ( ( Next == TAB )
                ||        ( Next == CARRIAGERETURN )
                ||        ( Next == LINEFEED ) )
                        Next = SPACE;

                if ( isprint(Next) ) *Data++ = Next;
        }

        *Data = EOS;

        fclose(fd);

        return TRUE;
}

/*------------------------------------------------------------------------*/

BOOLEAN FileExists(     char    *FileName )
{
        FILE *FD = fopen(FileName,"r");

        if ( FD == NULL ) return FALSE;

        fclose(FD);

        return TRUE;
}

/*------------------------------------------------------------------------*/

BOOLEAN SussFileName(   char    *GivenName,
                        char    *DefExtension )
{
static  char    Dir[MAX_PATH+ONE];
static  char    Name[MAX_PATH+ONE];
        char    *Slash;

        if ( FileExists(GivenName) ) return TRUE;

        strcpy(Name,GivenName);
        strcat(Name,DOTSTR);
        strcat(Name,DefExtension);

        if      ( FileExists(Name) )
        {
                strcpy(GivenName,Name);
                return TRUE;
        }

        if      ( ( GetModuleFileName(NULL,Dir,sizeof(Dir)-ONE) == ZERO )
        ||        ( (Slash=strrchr(Dir,BACKSLASH)) == NULL ) )
                return FALSE;

        Slash[ONE] = EOS;

        strcpy(Name,Dir);
        strcat(Name,GivenName);

        if      ( FileExists(Name) )
        {
                strcpy(GivenName,Name);
                return TRUE;
        }

        strcpy(Name,Dir);
        strcat(Name,GivenName);
        strcat(Name,DOTSTR);
        strcat(Name,DefExtension);

        if      ( FileExists(Name) )
        {
                strcpy(GivenName,Name);
                return TRUE;
        }

        return FALSE;
}

/*------------------------------------------------------------------------*/

BOOLEAN DoResponse(     char    *Cmd,
                        char    *OldCmd )
{
        int     Pos;
static  char    FileName[MAX_PATH+ONE];
static  char    FileText[MAX_COMMAND_LINE_LENGTH+ONE];

        DbgBoxTxt("DORESPONSE(s) OldCmd",OldCmd);

        strcpy(Cmd,OldCmd);

        while   ( (Pos=strpos(Cmd,ATSIGN)) >= ZERO )
        {
                Cmd += Pos;

                if      ( Cmd[ONE] == ATSIGN )
                {
                        strdel(Cmd,ZERO,ONE);
                        Cmd++;
                        continue;
                }

                if ( strpos(Cmd+ONE,ATSIGN) < ZERO ) return TRUE;

                strdel(Cmd,ZERO,ONE);
                Pos = strpos(Cmd,ATSIGN);
                strdel(Cmd,Pos,ONE);

                strncpy(FileName,Cmd,Pos);
                FileName[Pos] = EOS;

                strdel(Cmd,ZERO,Pos);

                if      ( !SussFileName(FileName,APPNAME) )
                {
                        BoxTxt("Couldn't find response file",FileName);
                        BoxErrGLE();
                        return FALSE;
                }

                if      ( !GetFileText(FileName,FileText) )
                {
                        BoxTxt("Couldn't read response file",FileName);
                        BoxErrGLE();
                        return FALSE;
                }

                strins(Cmd,FileText,ZERO);
        }

        DbgBoxTxt("DORESPONSE(e) Cmd",Cmd);

        return TRUE;
}

/*------------------------------------------------------------------------*/

BOOLEAN DoParams(       char    *Cmd,
                        char    *OldCmd )
{
        int Pos;
        char StrDelim;

        DbgBoxTxt("DOPARAMS(s) OldCmd",OldCmd);

        strcpy(Cmd,OldCmd);
        strcat(Cmd,SPACESTR);
        strcat(Cmd,SPACESTR);

        StripLeadChar(Cmd,SPACE);

        while   ( IsParam(Cmd) )
        {
                StrDelim = Cmd[ZERO];

                if      ( ( StrDelim == DOUBLEQUOTE )
                ||        ( StrDelim == SINGLEQUOTE ) )
                {
                        strdel(Cmd,ZERO,ONE);
                        Pos = strpos(Cmd,StrDelim);
                        if      ( Pos < ZERO )
                                Pos = strlen(Cmd) - TWO;
                        else
                                strdel(Cmd,Pos,ONE);
                }
                else
                        Pos = strpos(Cmd,SPACE);

                Cmd[Pos] = EOS;

                if ( !DoParam(Cmd) ) return FALSE;

                strcpy(Cmd,Cmd+Pos+ONE);
                StripLeadChar(Cmd,SPACE);
        }

        StripTrailChar(Cmd,SPACE);

        DbgBoxTxt("DOPARAMS(e) Cmd",Cmd);

        return TRUE;
}

/*------------------------------------------------------------------------*/

BOOL    FindFullProgName( BOOL  WantSearch,
                        char    *DefaultDir,
                        char    *CommandLine,
                        char    *FullProgName )
{
static  char      PartProgName[MAX_PATH+ONE];
        int       StartPos = ZERO;
        int       EndPos;
        HINSTANCE Result;
        char      StrDelim;

        DbgBoxTxt("FINDFULLPROGNAME(s) DefaultDir",DefaultDir);
        DbgBoxTxt("FINDFULLPROGNAME(s) CommandLine",CommandLine);

        StrDelim = CommandLine[ZERO];

        if      ( ( StrDelim == DOUBLEQUOTE ) || ( StrDelim == SINGLEQUOTE ) )
                StartPos = ONE,
                EndPos   = strpos(CommandLine+ONE,StrDelim) - ONE;
        else
        {
                EndPos = strpos(CommandLine,SPACE) - ONE;

                if      ( EndPos < ZERO )
                {
                        EndPos = strlen(CommandLine) - ONE;
                        CommandLine[EndPos+TWO] = EOS;
                }
        }

        if ( EndPos < StartPos ) return FALSE;

        strncpy(PartProgName,CommandLine+StartPos,EndPos-StartPos+ONE);
        PartProgName[EndPos-StartPos+ONE] = EOS;

        if      ( WantSearch )
        {
                Result = FindExecutable(PartProgName,DefaultDir,FullProgName);
                if ( (int) Result < 32 ) return FALSE;
        }
        else
                strcpy(FullProgName,PartProgName);

        strdel(CommandLine,ZERO,EndPos+TWO);

        DbgBoxTxt("FINDFULLPROGNAME(e) CommandLine",CommandLine);
        DbgBoxTxt("FINDFULLPROGNAME(e) FullProgName",FullProgName);

        return TRUE;
}

/*------------------------------------------------------------------------*/

int     ProcFinished(   HANDLE  RunningProcess )
{
        DWORD ExitCode;

        if      ( !GetExitCodeProcess(RunningProcess,&ExitCode) )
        {
                Box("Failed to get exit code");
                BoxErrGLE();
                return EXIT_CODE_BAD_EXITPROC;
        }

        if ( ExitCode == STILL_ACTIVE ) return STILL_ACTIVE;

        fflush(NULL);

        if      ( EchoOrders )
        {
                printf("\n" APPNAME ">> Last task exit code %d / 0x%X\n\n",
                        ExitCode,ExitCode);

                fflush(NULL);
        }

        return EXIT_CODE_SUCCESS;
}

/*------------------------------------------------------------------------*/

int     DoCommands(     char    *Orders )
{
static  char                    CmdLine[MAX_COMMAND_LINE_LENGTH+ONE];
static  char                    ComSpec[MAX_PATH+ONE];
        int                     Res;
        MSG                     Msg;
        STARTUPINFO             Start;
        PROCESS_INFORMATION     ProcInfo;
        SECURITY_ATTRIBUTES     Security;
        LPVOID                  EnvPtr;

        DbgBoxTxt("DOCOMMANDS(s) Orders",Orders);

        if      ( !IsPresent(StartDir)
        &&        ( GetCurrentDirectory(sizeof(StartDir),StartDir) == ZERO ) )
        {
                Box("Can't get current (start) directory");
                BoxErrGLE();
                return EXIT_CODE_BAD_DIR;
        }

        if      ( MakeConsole && !AllocConsole() )
        {
                Box("Can't create console window");
                BoxErrGLE();
                return EXIT_CODE_BAD_CONSOLE;
        }

        GetStartupInfo(&Start);

        if      ( MakeConsole
        &&      ( ShowState == SS_MAX
                  ||  SizX != ROGUE_INT ||  SizY != ROGUE_INT ) )
        {
                HANDLE ConHand = GetStdHandle(STD_OUTPUT_HANDLE);
                COORD  NewSize;

                if      ( ShowState == SS_MAX )
                {
                        NewSize = GetLargestConsoleWindowSize(ConHand);
                        if      ( NewSize.X == ZERO && NewSize.Y == ZERO )
                        {
                                Box("Can't get largest console window size");
                                BoxErrGLE();
                                return EXIT_CODE_BAD_CONGETSIZE;
                        }
                }
                else
                        NewSize.X = SizX, NewSize.Y = SizY;

                if      ( !SetConsoleScreenBufferSize(ConHand,NewSize) )
                {
                        Box("Can't re-size console window");
                        BoxErrGLE();
                        return EXIT_CODE_BAD_CONRESIZE;
                }
        }

        if      ( RedirectSTD )
        {
                Start.dwFlags    |= STARTF_USESTDHANDLES;
                Start.hStdInput  = GetStdHandle(STD_INPUT_HANDLE);
                Start.hStdOutput = GetStdHandle(STD_OUTPUT_HANDLE);
                Start.hStdError  = GetStdHandle(STD_ERROR_HANDLE);
        }

        if      ( ShowState != SS_NONE )
        {
                Start.dwFlags |= STARTF_USESHOWWINDOW;
                switch  ( ShowState )
                {
                case SS_MIN: Start.wShowWindow = SW_SHOWMINNOACTIVE; break;
                case SS_MAX: Start.wShowWindow = SW_SHOWMAXIMIZED;   break;
                case SS_RES: Start.wShowWindow = SW_SHOWDEFAULT;     break;
                }
        }

        if      ( ( CFollow ) || ( ( PosX != ROGUE_INT ) && ( PosY != ROGUE_INT ) ) )
        {
                Start.dwFlags |= STARTF_USEPOSITION;

                if      ( CFollow )
                {
                        POINT Pos;

                        if      ( GetCursorPos(&Pos) )
                                Start.dwX = Pos.x,
                                Start.dwY = Pos.y;
                }
                else
                        Start.dwX = PosX,
                        Start.dwY = PosY;
        }

        if      ( ( SizX != ROGUE_INT ) && ( SizY != ROGUE_INT ) )
        {
                Start.dwFlags   |= STARTF_USESIZE;
                Start.dwXSize   = SizX;
                Start.dwYSize   = SizY;
        }

        Security.nLength              = sizeof Security;
        Security.lpSecurityDescriptor = NULL;
        Security.bInheritHandle       = TRUE;

        StripLeadChar(Orders,SPACE);

        if      ( SplitCmd )
        {
                FindFullProgName(FALSE,StartDir,Orders,ComSpec);
                strcpy(CmdLine,Orders);
        }
        else
        if      ( !ForceShell && FindFullProgName(TRUE,StartDir,Orders,ComSpec) )
                sprintf(CmdLine,"%s %s",ComSpec,Orders);
        else
        {
               if (GetEnvironmentVariable(COMSPEC,ComSpec,sizeof ComSpec)==ZERO)
               {
                        BoxTxt("Command processor not found",COMSPEC);
                        return EXIT_CODE_NO_COMSPEC;
               }

               sprintf(CmdLine,"%s /c %s",ComSpec,Orders);
        }

        StripLeadChar(CmdLine,SPACE);
        StripTrailChar(CmdLine,SPACE);

        if      ( EchoOrders )
        {
                printf("\n" APPNAME ">> '%s> ",StartDir);
                if ( SplitCmd ) printf("[%s] ",ComSpec);
                printf("%s'\n\n",CmdLine);
        }

        fflush(NULL);

        if ( SplitCmd ) DbgBoxTxt("DOCOMMANDS() ComSpec",ComSpec);
        DbgBoxTxt("DOCOMMANDS() CmdLine",CmdLine);
        DbgBoxTxt("DOCOMMANDS() StartDir",StartDir);

        EnvPtr = PassEnvVar ? GetEnvironmentStrings() : NULL;

        if      ( !CreateProcess( SplitCmd ? ComSpec : NULL,
                                  CmdLine,
                                  &Security,            // Process security
                                  &Security,            // Thread security
                                  TRUE,                 // Inherit handles
                                  ZERO,                 // Creation flags
                                  EnvPtr,               // Environment
                                  StartDir,&Start,&ProcInfo ) )
        {
                BoxTxt("Process create failed",CmdLine);
                BoxErrGLE();
                return EXIT_CODE_BAD_EXECUTE;
        }

        if ( WaitTilDone == ROGUE_INT ) return EXIT_CODE_SUCCESS;

        while   ( (Res=ProcFinished(ProcInfo.hProcess)) == STILL_ACTIVE )
        {
                if ( WaitTilDone > ZERO ) Sleep(WaitTilDone);
                PeekMessage(&Msg,NULL,0,0,PM_NOREMOVE);
        }

        return Res;
}

/*------------------------------------------------------------------------*/

void
CALLBACK TimerProc(     HWND    hwnd,
                        UINT    uMsg,
                        UINT    uEvent,
                        DWORD   dwTime )
{
        Abandoned = TRUE;

        KillTimer(NULL,TimerHandle);
}

/*------------------------------------------------------------------------*/

BOOL    DoMenu(         void )
{
        HMENU   Menu;
        int     x,y,i;

        if      ( MenuItems <= ZERO )
        {
                Box("No menu items found");
                return FALSE;
        }

        Menu = CreatePopupMenu();

        if      ( Menu == NULL )
        {
                Box("Can't create popup menu");
                BoxErrGLE();
                return FALSE;
        }

        if ( BaseCmd == ROGUE_INT ) BaseCmd = WM_USER;

        for     ( i=ZERO ; i<MenuItems ; i++ )
                if    ( !AppendMenu(Menu,MF_STRING,BaseCmd+i,MenuItem[i].Name) )
                {
                      Box("Can't append to popup menu");
                      BoxErrGLE();
                      DestroyMenu(Menu);
                      return FALSE;
                }

        x = GetSystemMetrics(SM_CXSCREEN)/TWO;
        y = GetSystemMetrics(SM_CYSCREEN)/TWO;

        if      ( MFollow )
        {
                POINT Pos;

                if ( GetCursorPos(&Pos) ) x = Pos.x, y = Pos.y;
        }

        if      ( ( MPosX != ROGUE_INT ) && ( MPosY != ROGUE_INT ) )
                x = MPosX, y = MPosY;

        // If this next line is omitted, then one finds that one can't use the
        // menu (or escape from it) using the keyboard *if* the pick box, for
        // whatever reason, hasn't been displayed.

        SetActiveWindow(Root);

        if      ( !TrackPopupMenu(Menu,ZERO,x,y,ZERO,Root,NULL) )
        {
                Box("Can't track popup menu");
                BoxErrGLE();
                DestroyMenu(Menu);
                return FALSE;
        }

        if      ( !DestroyMenu(Menu) )
        {
                Box("Can't destroy popup menu");
                BoxErrGLE();
                return FALSE;
        }

        TimerHandle = SetTimer(NULL,ZERO,MENU_PERIOD,TimerProc);

        return TRUE;
}

/*------------------------------------------------------------------------*/

BOOL
CALLBACK DlgProcHelp(   HWND    hwnd,
                        UINT    iMsg,
                        WPARAM  wParam,
                        LPARAM  lParam)
{
        char *HelpText( void );

        switch  ( iMsg )
        {
        case WM_INITDIALOG:
                SetWindowText(hwnd,APPFULLMONTY);
                SetDlgItemText(hwnd,IDC_HELPTEXT,HelpText());
                return TRUE;

        case WM_COMMAND:
                if      ( HIWORD(wParam) == BN_CLICKED )
                {
                        EndDialog(hwnd,TRUE);
                        return TRUE;
                }
        }

        return FALSE;
}

/*------------------------------------------------------------------------*/

void    DoHelp(         HWND    hwnd )
{
        DialogBox(Inst,MAKEINTRESOURCE(IDD_HELP),hwnd,(DLGPROC)DlgProcHelp);
}

/*------------------------------------------------------------------------*/

LRESULT
CALLBACK WndProc(       HWND    hwnd,
                        UINT    iMsg,
                        WPARAM  wParam,
                        LPARAM  lParam)
{
        if      ( iMsg == WM_COMMAND )
        {
                PickCmd = LOWORD(wParam) - BaseCmd;
                DbgBoxInt("Menu item picked",PickCmd);
                return ZERO;
        }

        if      ( iMsg != RegisterWindowMessage(HELPMSGSTRING) )
                return DefWindowProc(hwnd,iMsg,wParam,lParam);

        DoHelp(hwnd);
        return ZERO;
}

/*------------------------------------------------------------------------*/

HWND    DoWindow(       HINSTANCE hInstance )
{
        WNDCLASS wndclass;
        HWND     Hwnd;
        int      x,y;

        wndclass.style         = ZERO;
        wndclass.lpfnWndProc   = WndProc;
        wndclass.cbClsExtra    = ZERO;
        wndclass.cbWndExtra    = ZERO;
        wndclass.hInstance     = Inst;
        wndclass.hIcon         = LoadIcon(hInstance,"PNR_ICON");
        wndclass.hCursor       = LoadCursor(NULL,IDC_ARROW);
        wndclass.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
        wndclass.lpszMenuName  = APPNAME " Dummy Window";
        wndclass.lpszClassName = APPANDVERANDDATE;

        RegisterClass(&wndclass);

        x = GetSystemMetrics(SM_CXSCREEN)/THREE;
        y = GetSystemMetrics(SM_CYSCREEN)/THREE;

        if      ( DFollow )
        {
                POINT Pos;

                if ( GetCursorPos(&Pos) ) x = Pos.x, y = Pos.y;
        }

        if      ( ( DPosX != ROGUE_INT ) && ( DPosY != ROGUE_INT ) )
                x = DPosX, y = DPosY;

        Hwnd = CreateWindow(APPANDVERANDDATE,APPANDVERANDDATE,WS_POPUP ,
                            x,y,50,50,NULL,NULL,Inst,NULL);

        if ( ( Hwnd != NULL ) && MakeVisible ) ShowWindow(Hwnd,SW_RESTORE);

        return Hwnd;
}

/*------------------------------------------------------------------------*/

BOOL    DoPickBox(      HWND    Hwnd )
{
        OPENFILENAME ofn;

static  char ft[MAX_PATH+ONE];

        ofn.lStructSize         = sizeof(ofn);
        ofn.hwndOwner           = Hwnd;
        ofn.hInstance           = ZERO;
        ofn.lpstrFilter         = Filters;
        ofn.lpstrCustomFilter   = NULL;
        ofn.nMaxCustFilter      = ZERO;
        ofn.nFilterIndex        = ZERO;
        ofn.lpstrFile           = FileMask;
        ofn.nMaxFile            = sizeof(FileMask) - ONE;
        ofn.lpstrFileTitle      = ft;
        ofn.nMaxFileTitle       = sizeof(ft) - ONE;
        ofn.lpstrInitialDir     = NULL;
        ofn.lpstrTitle          = Title;
        ofn.Flags               =   OFN_PATHMUSTEXIST
                                  | OFN_HIDEREADONLY;

        if ( !WantCreate ) ofn.Flags |= OFN_FILEMUSTEXIST;
        if ( WantHelp )    ofn.Flags |= OFN_SHOWHELP;
        if ( WantExp )     ofn.Flags |= OFN_EXPLORER;
        if ( WantMulti )   ofn.Flags |= ( OFN_EXPLORER | OFN_ALLOWMULTISELECT );

        ofn.nFileOffset         = ZERO;
        ofn.nFileExtension      = ZERO;
        ofn.lpstrDefExt         = NULL;
        ofn.lCustData           = ZERO;
        ofn.lpfnHook            = NULL;
        ofn.lpTemplateName      = NULL;

        return GetOpenFileName(&ofn);
}

/*------------------------------------------------------------------------*/

BOOLEAN OnlyFile(       char    *SingleFile,
                        char    *FileMask )
{
        HANDLE          Hand;
        WIN32_FIND_DATA Info;
static  char            FirstFile[MAX_PATH+ONE];
        BOOLEAN         MoreFiles;

        Hand = FindFirstFile(FileMask,&Info);

        if      ( Hand == INVALID_HANDLE_VALUE )
        {
                if      ( GetLastError() == ERROR_FILE_NOT_FOUND )
                        return TRUE;
                else
                {
                        BoxTxt("Cannot scan for first similar file",FileMask);
                        BoxErrGLE();
                        return FALSE;
                }
        }

        strcpy(FirstFile,Info.cFileName);

        MoreFiles = FindNextFile(Hand,&Info);

        if      ( !MoreFiles && ( GetLastError() != ERROR_NO_MORE_FILES ) )
        {
                BoxTxt("Cannot scan for similar files",FileMask);
                BoxErrGLE();
                return FALSE;
        }

        if      ( !FindClose(Hand) )
        {
                BoxTxt("Cannot close file scan handle",FileMask);
                BoxErrGLE();
                return FALSE;
        }

        if ( !MoreFiles ) strcpy(SingleFile,FirstFile);

        return TRUE;
}

/*------------------------------------------------------------------------*/

int     MessageLoops(   void )
{
        MSG  msg;

        while ( GetMessage(&msg,NULL,ZERO,ZERO) )
        {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
        }

        return msg.wParam;
}

/*------------------------------------------------------------------------*/

void    MessageLoop(    void )
{
        MSG  msg;

        if      ( GetMessage(&msg,NULL,ZERO,ZERO) )
        {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
        }
}

/*------------------------------------------------------------------------*/

BOOL    GetShortName(   char    *LongName,
                        char    *ShortName )
{
        int     RealLen;

        DbgBoxTxt("GETSHORTNAME(s) LongName",LongName);

        RealLen = GetShortPathName(LongName,ShortName,MAX_PATH+ONE);

        if      ( RealLen == ZERO )
        {
                BoxTxt("Cannot get short name for",LongName);
                BoxErrGLE();
                return FALSE;
        }

        if      ( RealLen > MAX_PATH )
        {
                BoxTxt("Short name buffer too small for",LongName);
                return FALSE;
        }

        DbgBoxTxt("GETSHORTNAME(e) ShortName",ShortName);

        return TRUE;
}

/*------------------------------------------------------------------------*/

int     DoSingleCmd(    char    *Command,
                        char    *FileName,
                        FILE    *Batch )
{
static  char    FileDriv[MAX_PATH+ONE];
static  char    FilePath[MAX_PATH+ONE];
static  char    FileRoot[MAX_PATH+ONE];
static  char    FileExtn[MAX_PATH+ONE];

        DbgBoxTxt("DOSINGLECMD(s) Command",Command);
        DbgBoxTxt("DOSINGLECMD(s) FileName",FileName);

        if      ( ShortFNames
        &&        !GetShortName(FileName,FileName) )
                return FALSE;

        ParseFileName(FileName,FileDriv,FilePath,FileRoot,FileExtn);

        if ( AutoQuote ) QuoteStr(FileName);

        DbgBoxTxt("DOSINGLECMD(1) Command",Command);
        if ( ExpandEnvAll(Command) == NULL ) return EXIT_CODE_PARAM_CANCEL;
        DbgBoxTxt("DOSINGLECMD(2) Command",Command);
        ExpandStr(Command,FileName,FileDriv,FilePath,FileRoot,FileExtn);
        DbgBoxTxt("DOSINGLECMD(3) Command",Command);

        if      ( Batch != NULL )
        {
                fprintf(Batch,"%s\n",Command);
                return EXIT_CODE_SUCCESS;
        }

        if ( AutoDir && !IsPresent(StartDir) ) strcpy(StartDir,FilePath);

        return DoCommands(Command);
}

/*------------------------------------------------------------------------*/

int     DoMultiCmds(    char    *Command,
                        char    *FileName,
                        FILE    *Batch )
{
        char    *DirName = FileName;
static  char    NextFile[MAX_PATH+ONE];
static  char    NextCommand[MAX_COMMAND_LINE_LENGTH+ONE];
        int     Res;

        DbgBoxTxt("DOMULTICMDS(s) Command",Command);
        DbgBoxTxt("DOMULTICMDS(s) FileName",FileName);

        FileName += ( strlen(FileName) + ONE );

        while   ( *FileName != EOS )
        {
                DbgBoxTxt("DOMULTICMDS() FileName",FileName);

                strcpy(NextCommand,Command);

                strcpy(NextFile,DirName);
                strcat(NextFile,BACKSLASHSTR);
                strcat(NextFile,FileName);

                Res = DoSingleCmd(NextCommand,NextFile,Batch);

                if ( Res != EXIT_CODE_SUCCESS ) return Res;

                FileName += ( strlen(FileName) + ONE );

                if      ( *FileName != EOS && ( LaunchGap > ZERO ) )
                        Sleep(LaunchGap);
        }

        return EXIT_CODE_SUCCESS;
}

/*------------------------------------------------------------------------*/

int     DoMultiCmd(     char    *Command,
                        char    *FileName,
                        FILE    *Batch )
{
        char    *DirName = FileName;
        BOOL    WantDir  = TRUE;

static  char    NextFile[MAX_PATH+ONE];
static  char    AllFiles[MAX_COMMAND_LINE_LENGTH+ONE];

static  char    FileDriv[MAX_PATH+ONE];
static  char    FilePath[MAX_PATH+ONE];
static  char    FileRoot[MAX_PATH+ONE];
static  char    FileExtn[MAX_PATH+ONE];

static  char    FileDriv2[MAX_PATH+ONE];
static  char    FilePath2[MAX_PATH+ONE];
static  char    FileRoot2[MAX_PATH+ONE];
static  char    FileExtn2[MAX_PATH+ONE];

        DbgBoxTxt("DOMULTICMD(s) Command",Command);
        DbgBoxTxt("DOMULTICMD(s) FileName",FileName);

        if      ( AutoDir && !IsPresent(StartDir) )
        {
                strcpy(StartDir,DirName);
                WantDir = FALSE;
        }

        FileName += ( strlen(FileName) + ONE );

        strcpy(NextFile,DirName);
        strcat(NextFile,BACKSLASHSTR);
        strcat(NextFile,FileName);

        if      ( ShortFNames
        &&        !GetShortName(NextFile,NextFile) )
                return EXIT_CODE_BAD_SHORTNAME;

        ParseFileName(NextFile,FileDriv,FilePath,FileRoot,FileExtn);

        strcpy(AllFiles,NILSTR);

        while   ( *FileName != EOS )
        {
                DbgBoxTxt("DOMULTICMD() FileName",FileName);

                if      ( WantDir )
                {
                        strcpy(NextFile,DirName);
                        strcat(NextFile,BACKSLASHSTR);
                        strcat(NextFile,FileName);
                }
                else
                        strcpy(NextFile,FileName);

                if      ( ShortFNames
                &&        !GetShortName(NextFile,NextFile) )
                        return EXIT_CODE_BAD_SHORTNAME;

                ParseFileName(NextFile,FileDriv2,FilePath2,FileRoot2,FileExtn2);

                if ( !IsSame(FileRoot,FileRoot2) ) strcpy(FileRoot,NILSTR);
                if ( !IsSame(FileExtn,FileExtn2) ) strcpy(FileExtn,NILSTR);

                if ( IsPresent(AllFiles) ) strcat(AllFiles,SPACESTR);

                if ( AutoQuote ) QuoteStr(NextFile);

                strcat(AllFiles,NextFile);

                FileName += ( strlen(FileName) + ONE );
        }

        DbgBoxTxt("DOMULTICMD(1) Command",Command);
        if ( ExpandEnvAll(Command) == NULL ) return EXIT_CODE_PARAM_CANCEL;
        DbgBoxTxt("DOMULTICMD(2) Command",Command);
        ExpandStr(Command,AllFiles,FileDriv,FilePath,FileRoot,FileExtn);
        DbgBoxTxt("DOMULTICMD(3) Command",Command);

        if      ( Batch != NULL )
        {
                fprintf(Batch,"%s\n",Command);
                return EXIT_CODE_SUCCESS;
        }

        return DoCommands(Command);
}

/*------------------------------------------------------------------------*/

int WINAPI WinMain(     HINSTANCE hInstance,
                        HINSTANCE hPrevInstance,
                        PSTR      szCmdLine,
                        int       iCmdShow )
{
static  char    CmdLine1[MAX_COMMAND_LINE_LENGTH+ONE];
static  char    CmdLine2[MAX_COMMAND_LINE_LENGTH+ONE];

static  char    SaveCmdLine2[MAX_COMMAND_LINE_LENGTH+ONE];
static  char    SaveFileMask[MAX_PATH+ONE];
static  char    SaveStartDir[MAX_PATH+ONE];
static  char    SingleFile[MAX_PATH+ONE];

        int     Res;
        FILE    *Batch = NULL;
        BOOLEAN Done = FALSE;

        Inst = hInstance;

        for     ( ParamThis=ZERO ; ParamThis<PARAMS ; ParamThis++ )
                strcpy(ParamText[ParamThis],NILSTR);

        strcpy(FileMask,NILSTR);
        strcpy(OFileName,NILSTR);
        strcpy(Title,APPFULLMONTY);
        strcpy(StartDir,NILSTR);
        strcpy(SingleFile,NILSTR);

        if      ( !DoResponse(CmdLine1,szCmdLine) )
                return EXIT_CODE_BAD_RESPONSE;

        if      ( !DoParams(CmdLine2,CmdLine1) )
                return EXIT_CODE_BAD_PARAMS;

        DbgBoxTxt("WINMAIN(s) szCmdLine",szCmdLine);
        DbgBoxTxt("WINMAIN(s) CmdLine1",CmdLine1);

        strcpy(SaveCmdLine2,CmdLine2);
        strcpy(SaveFileMask,FileMask);
        strcpy(SaveStartDir,StartDir);

        while   ( !Done )
        {
                memset((void*)FileMask,ZERO,sizeof(FileMask));

                strcpy(CmdLine2,SaveCmdLine2);
                strcpy(FileMask,SaveFileMask);
                strcpy(StartDir,SaveStartDir);

                DbgBoxTxt("WINMAIN() CmdLine2",CmdLine2);
                DbgBoxTxt("WINMAIN() FileMask",FileMask);
                DbgBoxTxt("WINMAIN() StartDir",StartDir);

                if      ( (Root=DoWindow(hInstance)) == NULL )
                {
                        Box("Window create failed");
                        BoxErrGLE();
                        return EXIT_CODE_BAD_WINDOW;
                }

                if      ( !IsPresent(CmdLine2) && ( MenuItems <= ZERO ) )
                {
                        DoHelp(Root);
                        return EXIT_CODE_SUCCESS;
                }

                if      ( !InhibitPick )
                {
                        if      ( PickSingle
                        &&      !OnlyFile(SingleFile,FileMask) )
                                return EXIT_CODE_BAD_FILEFIND;

                        if      ( IsPresent(SingleFile) )
                                strcpy(FileMask,SingleFile);
                        else
                                if      ( !DoPickBox(Root) )
                                        return EXIT_CODE_BAD_SELECTION;
                }

                if      ( IsPresent(OFileName) )
                {
                        Batch = fopen(OFileName,"w");

                        if      ( Batch == NULL )
                        {
                           BoxTxt("Batch file create failed",OFileName);
                           BoxErrGLE();
                           return EXIT_CODE_BAD_FOPEN;
                        }
                }

                if      ( MenuItems > ZERO )
                {
                        Abandoned = FALSE;
                        PickCmd   = ROGUE_INT;

                        if ( !DoMenu() ) return EXIT_CODE_BAD_MENU;

                        while   ( ( PickCmd == ROGUE_INT ) && !Abandoned )
                                MessageLoop();

                        if ( Abandoned ) return EXIT_CODE_BAD_SELECTION;

                        strcpy(CmdLine2,MenuItem[PickCmd].Work);
                }

                if      ( !WantMulti
                ||        ( *(FileMask+strlen(FileMask)+ONE) == EOS ) )
                        Res = DoSingleCmd(CmdLine2,FileMask,Batch);
                else
                if      ( OneMultiCmd )
                        Res = DoMultiCmd(CmdLine2,FileMask,Batch);
                else
                        Res = DoMultiCmds(CmdLine2,FileMask,Batch);

                if ( IsPresent(OFileName) ) fclose(Batch);

/* TSE                 if      ( ( Res == EXIT_CODE_SUCCESS ) TSE */
/* TSE                 &&        !DestroyWindow(Root) ) TSE */
                if      ( !DestroyWindow(Root) )
                {
                        Box("Window destroy failed");
                        BoxErrGLE();
                        return EXIT_CODE_BAD_DESTROY;
                }

                Done = ( !WantRepeat ) || ( Res != EXIT_CODE_SUCCESS );
        }

        return Res;
}

/*========================================================================*/

static  char    HelpTextBuffer[] =
{
"\r\n"
"		PNR by Mark Steven Davis (G0KHB)\r\n"
"\r\n"
"DIALOGUE PARAMETERS:\r\n"
"\r\n"
"-DA		Sets dialogue box position from current cursor postion.\r\n"
"-DC		Allows the user to nominate a file that does *not* (yet) exist.\r\n"
"-DF<filters>	Sets the exclamation-mark delimited \"filters\" for the dialogue.\r\n"
"-DH		Turns off the help button in the pick dialogue.\r\n"
"-DI		Makes the \"hidden\" pick box parent window visible.\r\n"
"-DM<masks>	Specifies the set of files that will appear in the pick box.\r\n"
"-DO<filename>	Causes commands to be sent to the named batch file.\r\n"
"-DP<x>,<y>	Allows the user to sets the pick dialogue screen location.\r\n"
"-DR		Displays pick boxes and executes commands until canceled.\r\n"
"-DS		By-passes dialogue box if only *one* match of single file type.\r\n"
"-DT<string>	Sets the caption in the title of the pick box.\r\n"
"-DU		Multi-select enable - more than one file can be picked and run.\r\n"
"-DV<name>=<val>	Parent environment variable <name> is set to <val>.\r\n"
"-DW[<ms>]	Wait <ms> milliseconds between each command-line launch.\r\n"
"-DX		Inhibits the displaying of the pick box - used for side-effects.\r\n"
"\r\n"
"CHILD PROCESS PARAMETERS:\r\n"
"\r\n"
"-CA		Attempts to set child window position from cursor postion.\r\n"
"-CC		Splits command-line and passes program and parameters separately.\r\n"
"-CD<directory>	Specifies the directory in which the commands will be started.\r\n"
"-CE		If used, the expanded commands will be sent to standard output.\r\n"
"-CG		Executes a multi-file select command using a *single* command shell\r\n"
"-CH		Forces the passing of short-form file names to the child.\r\n"
"-CI		Sets the \"inherit handles\" fields in the STARTUPINFO structure.\r\n"
"-C{M,N,R}	Makes PNR try to alter the initial display state of it's child.\r\n"
"-CO		Run the child process in a console window.\r\n"
"-CP<x>,<y>	PNR will attempt to set the child window position as (x,y).\r\n"
"-CQ		Disables the automatic quoting of file names that contain a space.\r\n"
"-CS		Forces an MSDOS command shell to be invoked to run the child.\r\n"
"-CU[<ms>]	Forces PNR to wait (sleep <ms>) until child has terminated.\r\n"
"-CV<name>=<val>	Child (and parent) environment variable <name> is set to <val>.\r\n"
"-CY		Disables automatic deduction of the start directory.\r\n"
"-CZ<x>,<y>	Attempt to use (x,y) as the child's window size.\r\n"
"\r\n"
"MENU PARAMETERS:\r\n"
"\r\n"
"-MA		Sets menu position from current cursor postion.\r\n"
"-MB<wm_base>	Numbers the first menu WM_COMMAND at number <wm_base>.\r\n"
"-MC<command>	Fill in the current menu item with command <command>.\r\n"
"-MN<name>	Fill in the current menu description text with <name>.\r\n"
"-MP<x>,<y>	PNR will attempt to set the menu position as (x,y).\r\n"
"\r\n"
"MISCELLANEOUS PARAMETERS:\r\n"
"\r\n"
"-XX		Send diagnostic messages to the Windows debug output stream.\r\n"
"@<filename>@	Takes text from a \"response\" file; to escape, use \"@@\".\r\n"
"\r\n"
"MINI-TUTORIAL:\r\n"
"\r\n"
"PNR is a program \"launcher\".  It allows the user to pick files, using a\r\n"
"standard \"file open dialogue\", and then execute a command after replacing\r\n"
"all occurrences of a special string by the name of the file picked.  In a\r\n"
"sense, PNR is a Windows equivalent of small MSdog batch files that start\r\n"
"simple processes.\r\n"
"\r\n"
"For example, under MSdog, one could have a batch file, COOK.BAT say, that\r\n"
"starts a compilation process.  Assuming there was a project called PROBE,\r\n"
"one could compile PROBE by using \"COOK PROBE\".  You can still do this, using\r\n"
"a DOS box, under Windows, of course.  However, many programmers prefer to\r\n"
"decorate their desktops with short-cut icons that essentially do the same\r\n"
"thing.  However, the command-lines launched by these icons are *fixed*.  It\r\n"
"would be to be able to pick a file and then have the icon launch a task\r\n"
"using the picked file as a \"target\".  This is where PNR comes in...\r\n"
"\r\n"
"For a concrete example, let's use MAKE.  We will assume that you wish to\r\n"
"have a short-cut that will allow you to pick a .MAK file from a directory\r\n"
"and then launch MAKE using that .MAK file as a target.  Attaching the\r\n"
"following command to a short-cut will have to desired effect:\r\n"
"\r\n"
"		PNR -DM*.MAK MAKE -F %1\r\n"
"\r\n"
"In this instance, when it is run, PNR will process all parameters on the\r\n"
"command line until it comes to the first one that does *not* start with a\r\n"
"minus-sign or slash - in this case the parameter \"MAKE\".  Then, based upon\r\n"
"the parameters that it *has* processed, it will display the file open\r\n"
"dialogue. For our example, it will have processed the \"-DM*.MAK\" and will,\r\n"
"therefore, use the default file mask of \"*.MAK\".\r\n"
"\r\n"
"When a file is picked, say PROBE.MAK, PNR will replace all occurrences of\r\n"
"\"%1\" in the rest of the command-line by the *full* name of the picked file.\r\n"
"In our example, it will replace the %1 in \"MAKE -F %1\" to give, say, \"MAKE\r\n"
"-F C:\\WORK\\C\\TC\\PROBE.MAK\".  Finally, it would submit this command-line\r\n"
"to the system for execution.\r\n"
"\r\n"
"Instead of using \"%1\" to obatin the full file name, *parts* of the file name\r\n"
"can be extracted:\r\n"
"\r\n"
"	%1 Full drive, path and file name.\r\n"
"	%2 Drive letter and colon.\r\n"
"	%3 Path excluding drive and file name.\r\n"
"	%4 File \"root\".\r\n"
"	%5 File extension.\r\n"
"\r\n"
"For example, if the picked file was \"C:\\1989\\AUGUST\\TEST.DBF\" the parts\r\n"
"would be extracted like this:\r\n"
"\r\n"
"	%1 \"C:\\1989\\AUGUST\\TEST.DBF\"\r\n"
"	%2 \"C\"\r\n"
"	%3 \"\\1989\\AUGUST\"\r\n"
"	%4 \"TEST\"\r\n"
"	%5 \"DBF\"\r\n"
"\r\n"
"There are parameters that allow the user to control the open file dialogue's\r\n"
"title bar and position.  You can also affect the appearance of the process\r\n"
"that is *started* by PNR; the size, position and start directory of the\r\n"
"process can all be changed.  Note, however, that some of the parameters may\r\n"
"have no effect depending upon the process that you are starting.  (That is,\r\n"
"whether it's a DOS box, Windows program, or whatever.)\r\n"
"\r\n"
"You can also make PNR either terminate as soon as the child process is\r\n"
"running or hang around until the child process terminates.  Note that PNR\r\n"
"makes the child inherit PNR's parent process handles and, therefore, can be\r\n"
"used as a \"pipe\".\r\n"
"\r\n"
"File \"filters\" can be used via the -DF parameter.  The filters should\r\n"
"comprise pairs of strings; the first string is used as the desciptive text\r\n"
"for the following file mask.  Multiple masks for a filter can be separated\r\n"
"by a semi-colon character.  The separate strings are separated by\r\n"
"exclamation-marks.  Note that if you want spaces to appear in the\r\n"
"description fields, you'll have to use the quotes form of parameter.\r\n"
"\r\n"
"This is an example of a relatively complicated -DF command:\r\n"
"\r\n"
"	\"-DF Forth Sources ! *.fth ! C Sources ! *.c;*.h\"\r\n"
"\r\n"
"Note that all -DF commands are combined into *one* filter lists, so the\r\n"
"previous example could have equally well be specified as:\r\n"
"\r\n"
"	\"-DF Forth Sources ! *.fth\" \"-DF C Sources ! *.c;*.h\"\r\n"
"\r\n"
"When the -DS parameter is specified and PNR has been given a *single* file\r\n"
"type to look for, the pick diagloue box will be inhibited if only *one* file\r\n"
"matches the mask.  In this case, PNR will go straight to the spawning of the\r\n"
"required command-line.  Note that this \"short-circuiting\" will *not* work if\r\n"
"PNR is presented with a set of masks concatenated using semi-colons.\r\n"
"\r\n"
"PARAMETER SYNTAX:\r\n"
"\r\n"
"You can following any switch parameter setting by a minus sign ('-') or a\r\n"
"plus sign ('+').  In the case of the former, the switch will be set to\r\n"
"\"false\" and in the case of the latter to \"true\".  Otherwise, the switch is\r\n"
"*toggled*.  As an example, the following will enable, disable and toggle\r\n"
"debugging respectively. (Note that as debugging defaults \"false\", the last\r\n"
"example will, in fact, *enable* debugging.)\r\n"
"\r\n"
"	-XX-\r\n"
"	-XX+\r\n"
"	-XX\r\n"
"\r\n"
"USING MENUS:\r\n"
"\r\n"
"So far, we've only given examples of files being picked and a single command\r\n"
"being executed.  However, PNR has the capability to allow the user to pick a\r\n"
"command from a popup menu.  For example, let's assume that you wanted to either\r\n"
"edit a file using NOTEPAD or using whatever you'd associated with the relevant\r\n"
"file type. This would suffice:\r\n"
"\r\n"
"	PNR \"-mnNotePad File\" \"-mcnotepad %1\" \"-mnStart File\" \"-mcstart %1\"\r\n"
"\r\n"
"Note the use of double quotes around parameters that contain any space\r\n"
"characters.\r\n"
"\r\n"
"RESPONSE FILES:\r\n"
"\r\n"
"The \"response\" file facility is invoked using the form:\r\n"
"\r\n"
"	@<filename>@\r\n"
"\r\n"
"This causes PNR to take parameters from a text file.  The program will\r\n"
"search in current directory for the cited file.  If it does not exist, the\r\n"
"program will append the default file extension of \".PNR\" and try again.  If\r\n"
"it *still* can't find the file, it will try the given file name in the\r\n"
"program's .EXE directory; finally, it will try the .EXE directory after\r\n"
"applying the default extension.\r\n"
"\r\n"
"This means that if you create a file called, say, TEXT.PNR in the same\r\n"
"location as the .EXE file, the command \"@TEXT@\" is sufficient to allow PNR\r\n"
"to locate and load the file; this allows command-lines to be kept much\r\n"
"shorter.\r\n"
"\r\n"
"USER-ENTERED PARAMETERS:\r\n"
"\r\n"
"If PNR treats environment variable names of the following form (where \"<n>\"\r\n"
"is one of the digits 0..9) in a special manner:\r\n"
"\r\n"
"        %PARAM<n>%\r\n"
"\r\n"
"In these cases, PNR will display a dialogue box that prompts the user for a\r\n"
"parameter.  For example, in order to invoke an editor (in this case, \"e\"), having\r\n"
"asked the user to enter two file names, this could be used:\r\n"
"\r\n"
"        PNR E -DX %PARAM1% %PARAM2%\r\n"
"\r\n"
"Note that if this is typed directly at a command shell prompt, it may have to be\r\n"
"entered as:\r\n"
"\r\n"
"        PNR E -DX %%PARAM1%% %%PARAM2%%\r\n"
"\r\n"
"This is because the shell will typically try to expand the special parameter\r\n"
"sequences as environment variables.  (We have, of course, tried to *mimic*\r\n"
"this functionality.)\r\n"
"\r\n"
"CHANGE LOG:\r\n"
"\r\n"
"2.1.015\r\n"
"\r\n"
"Added the multi-select single-command facility. In the past, when a\r\n"
"multi-select was done, PNR would start a separate command shell for *each*\r\n"
"file slected.  This new alternative allows you to start a single command\r\n"
"shell placing *all* the selected files in one \"%1\" parameter. The number of\r\n"
"command-shells launched is controlled via the \"-CG\" parameter.\r\n"
"\r\n"
"Earlier versions of PNR would allow you to set the \"current directory\" for\r\n"
"the command(s) to be in.  If the this \"start directory\" wasn't set\r\n"
"explicitly, then the \"current\" directory was used.  There is a further\r\n"
"option now available: if a *multi-file* select is done and *no* start\r\n"
"directory is dictated, then PNR will deduce the start directory from the\r\n"
"list of file names and *not* prepend each file name with a path.  This has\r\n"
"the benficial effect of making multi-selected, single command-shell command\r\n"
"lines *much* shorter.  This automatic start directory detection is\r\n"
"controlled via the \"-CY\" option.\r\n"
"\r\n"
"PNR will now *automatically* \"quote\" file names that contain spaces.  Note\r\n"
"that it will *not* quote any parameter other than \"%1\" - the drive, path,\r\n"
"root and extension are all left alone since PNR doesn't know what you\r\n"
"intend to do with these components. This action is controlled by the \"-CQ\"\r\n"
"parameter.\r\n"
"\r\n"
"As detailed above, parameters can now be toggled (\"-CY\") switched on\r\n"
"(\"-CY+\") or switched off (\"-CY-\").  Past versions of PNR only allowed the\r\n"
"toggling action so it was more important to know what the *default* state of\r\n"
"a switch was.\r\n"
"\r\n"
"You can now force PNR to use *short-form* file names when expanding a\r\n"
"command line.  The controlling parameter is \"-CH\".\r\n"
"\r\n"
"2.1.029\r\n"
"\r\n"
"Slight alteration to the way that the %2, %3, %4 and %5 command-line\r\n"
"parameters were handled when multi-select, *single* command-line shells\r\n"
"were run.\r\n"
"\r\n"
"2.2.009\r\n"
"\r\n"
"(a) Added the menu generation facility.  The -MN and -MC parameters can now\r\n"
"be used to make PNR create a menu.  The idea is that, having picked some\r\n"
"files, the user can choose what action to execute from a popup menu\r\n"
"that is created on the fly.\r\n"
"\r\n"
"(b) The parsing of response files was improved so that \"white space\"\r\n"
"(carriage-returns, line-feeds and tab characters) are now converted to\r\n"
"spaces before the text is included into the command-line.  As a result, the\r\n"
"text in response files may be laid out to be much more readable.\r\n"
"\r\n"
"2.2.011\r\n"
"\r\n"
"Increased the maximum number of menu items from 20 to 40.\r\n"
"\r\n"
"2.3.000\r\n"
"\r\n"
"Added the \"-DC\" parameter that let's you nominate a file name that\r\n"
"*doesn't* exist.  Also, added the use of PNR's very own icon.\r\n"
"\r\n"
"2.3.002\r\n"
"\r\n"
"Added the \"-DW[<ms>]\" parameter.  When specified, PNR will wait the required\r\n"
"number of milli-seconds between each launch of a command-line.  This waiting\r\n"
"will only be done if multiple files are selected and single command-line\r\n"
"launch is *not* selected.\r\n"
"\r\n"
"2.3.010\r\n"
"\r\n"
"Re-directed all debugging text messages to the Windows diagnostic stream using\r\n"
"the API call, OutputDebugString().  This is a much neater way of displaying the\r\n"
"diagnostic text but means that you *must* run a program that can *view* the\r\n"
"debug output stream. (Try searching the web for DBGVIEW, for example.) Unless\r\n"
"you are running such a utility, you will see *nothing* when the \"-XX\" option\r\n"
"is invoked.\r\n"
"\r\n"
"2.3.015\r\n"
"\r\n"
"Fixed a small bug related to PNR erroneously treating the results of the pick\r\n"
"box as if more than one file had been picked when this wasn't, if fact, the\r\n"
"case.\r\n"
"\r\n"
"2.4.000\r\n"
"\r\n"
"(a) Added the %PARAM<number>% facility.\r\n"
"\r\n"
"(b) Fixed the bug whereby multiple percent signs were sometimes required in\r\n"
"order to get a *single* percent sign to be processed.  This still *is* sometimes\r\n"
"necessary (when a PNR command-line has been placed in a .BAT file, for example),\r\n"
"but in many instances, PNR was simply applying the percent sign \"escapement\"\r\n"
"rules too frequently.  Note that the fixing of this bug means that *existing*\r\n"
"PNR command-lines may not work correctly!\r\n"
"\r\n"
"(c) Removed a couple of DEFWINDOWPROC() calls from within *dialogue* call-back\r\n"
"procedures because the WIN32 API help file says that you should *not* do this.\r\n"
"\r\n"
"2.5.000\r\n"
"\r\n"
"If the pick box, for whatever reason, wasn't displayed, then any following\r\n"
"menu couldn't be used from the keyboard. This has been fixed.\r\n"
"\r\n"
"2.6.000\r\n"
"\r\n"
"Added the -DA, -CA and -MA options.\r\n"
"\r\n"
"PNR is a ChAoS Utility - the ChAoS Utilities home page is www.glod.net\r\n"
};

char *HelpText( void ) { return HelpTextBuffer; }
